#include "DataManager.h"
#include "presds.h"

DataManager *DataManager::dmInstance=NULL;
struct server_conf* DataManager::sconf=NULL;

//////////////////////////////////////
// Added for RLS
/////////////////////////////////////
/**
 * For the mapping list subscriber and list uri
 */
map<string, string> DataManager::list_sub_record;

/**
 * For the mapping rls presentity and rls presentity object
 */
map<string, RLSPresentityDS*> DataManager::rls_pres_record;

/**
 * For the mapping presentity and resource list
 */
map<string, vector<string> > DataManager::rls_resources_record;



/**
 * per presentity record stored globally.
 */
map<string,PresentityDS*> DataManager::pres_record;

/**
 * Recursive lock for global presentity table.
 */
pthread_mutex_t DataManager::lock_pres_record= PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;

/**
 * Datamanager class.
 * Provides access to presentity and watcher records
 */

//TODO: Make the constructor private
DataManager ::DataManager()
{
  pthread_mutex_init(&lock_pres_record,NULL);
  list_sub_record["yb2149"] = "yb2149-buddies";
}

DataManager * DataManager::getInstance()
{
	if(dmInstance==NULL)
		dmInstance= new DataManager();
	
	return dmInstance;
}

DataManager ::~DataManager()
{
	pthread_mutex_destroy(&lock_pres_record);
}


///////////////////////////////////
// Added for RLS
///////////////////////////////////
/**
 * Returns name of the resourc list based on the presentity
 */
string DataManager::getTheResourceList(string presentity)
{
  string listID = "default";
  
  //Assumed there is only one List for one user
  map<string, string> ::iterator iter; 
  for(iter = list_sub_record.begin(); iter != list_sub_record.end(); iter++) {
    if (presentity == iter->first){
      listID = iter->second;
    }
  }
  
  return listID;
}

///////////////////////////////////
// Added for RLS
///////////////////////////////////
void DataManager::setTheResourceList(string presentity, string listName){
  list_sub_record[presentity] = listName;
}

///////////////////////////////////
// Added for RLS
///////////////////////////////////
/**
 * Returns resources in the list 
 */
vector<string> DataManager::getResources(string presentity)
{
  vector<string> resources;
  
  map<string, vector<string> > ::iterator iter; 
  for(iter=rls_resources_record.begin(); iter!=rls_resources_record.end(); iter++) {
    if (presentity == iter->first){
      resources = iter->second;
    }
  }
  
  return resources;
}

///////////////////////////////////
// Added for RLS
///////////////////////////////////
void DataManager::setResources(string presentity, vector<string> resources)
{
  rls_resources_record[presentity]= resources;
}



/////////////////////
//Added for RLS
/////////////////////
/**
 * Creates RLS Presentity based on the presentity
 */
void* DataManager::createRLSPresentity(string presentity)
{
	
  RLSPresentityDS *rpds = new RLSPresentityDS(presentity);
  rls_pres_record[presentity]= rpds;
  
  return rpds;
}

////////////////////
//Added for RLS
////////////////////
RLSPresentityDS* DataManager::getRLSPresentityDS(string presentity)
{
  RLSPresentityDS* ret = NULL;
  map<string, RLSPresentityDS*> ::iterator iter;

  for(iter = rls_pres_record.begin(); iter != rls_pres_record.end(); iter++) {
    if (presentity == iter->first){
      ret = iter->second;
    }
  }

  return ret;
}

////////////////////
//Added for RLS
////////////////////
int DataManager::removeRLSPresentity(string presentity)
{ 
  RLSPresentityDS* rpds = getRLSPresentityDS(presentity);
  if(rpds == NULL){
    return -1;
  }
  rls_pres_record.erase(presentity);
  delete rpds;

  return 1;
}


void DataManager ::setServerConf(struct server_conf *s)
{
	sconf =s;
}
struct server_conf*  DataManager :: getServerConf()
{
	return sconf;
}

/**
 * Creates an object corresponding to a presen object,
 * Create a DataContainer and A lock for Container for the presentity,
 */
void * DataManager::createPresentity(const string & presentity)
{
	
  map<string,pthread_mutex_t*> ::iterator pos;
  
  Lock(lock_pres_record,"Locking to creating presentity");
  PresentityDS *pds =NULL;
  pds = getPresentityDS(presentity);

  //if(ifExists(presentity)){}
  if(pds!=NULL){
    debug(DEBUG_IMPP, "DataManager::createPresentity", 
	  "presentity %s exists\n",presentity.c_str());
    UnLock(lock_pres_record,"UnLocking after creating presentity");
    return pds;
    
  }
  debug(DEBUG_IMPP, "DataManager::createPresentity", 
	"presentity %s doesnt exist.creating presentity\n",
	presentity.c_str());
  
  
  pds = new PresentityDS(presentity);
  pres_record[presentity]= pds;
  
  UnLock(lock_pres_record,"UnLocking after creating presentity");
  return pds;
}



/** 
 * Checks and tells if presentity exists in global presentity table
 */
bool DataManager::ifExists(const string &  presentity)
{


        map<string,PresentityDS*> ::iterator pos;

	Lock(lock_pres_record,"Locking to search for prsentity");

	pos = pres_record.find(presentity);

	bool ret = false;

	
	if (pos == pres_record.end()) {
		debug(DEBUG_IMPP, "DataManager::ifExists",
		"presentity %s not found\n",presentity.c_str());
		ret = false;
	}
	else {
		debug(DEBUG_IMPP, "DataManager::ifExists",
		"presentity %s found\n",presentity.c_str());
		ret=  true;
	}
	
	UnLock(lock_pres_record,"UnLocking after searching for presentity");
	
	return ret;

}

PresentityDS* DataManager::getPresentityDS(const string &  presentity)
{
        PresentityDS* ret =NULL;
        map<string,PresentityDS*> ::iterator pos;
	debug(DEBUG_IMPP, "DataManager::getPresentityDS",
	"searching for %s\n",presentity.c_str());
	Lock(lock_pres_record,"Locking to search for prsentity");
	pos = pres_record.find(presentity);
	if (pos != pres_record.end()) {
		debug(DEBUG_IMPP, "DataManager::getPresentityDS",
		"found record for Presentity %s\n",presentity.c_str());
		ret=pos->second;
	}else {
		debug(DEBUG_IMPP, "DataManager::getPresentityDS",
	        "not found record for Presentity %s\n",presentity.c_str());
    	}
	UnLock(lock_pres_record,"UnLocking after searching for presentity");
	return ret;

}

int DataManager:: removePresentity(const string & pres)
{ 
	debug(DEBUG_IMPP, "DataManager::removePresentity",
	"Removing presentity %s\n",pres.c_str());

        PresentityDS* ret =NULL;

	// Delete the Container
        map<string,PresentityDS*> ::iterator pos;

	Lock(lock_pres_record,"Locking to remove prsentity");


	pos = pres_record.find(pres);
	if (pos != pres_record.end()) {
		debug(DEBUG_IMPP, "DataManager::removePresentity",
		"found record for %s\n",(pos->first).c_str());
	        ret=pos->second;
	}else {
		debug(DEBUG_IMPP, "DataManager::removePresentity",
		"Not found record for %s\n",(pos->first).c_str());
		UnLock(lock_pres_record,"UnLocking,failed removing presentity");
        	return -1;
   	}
	pres_record.erase(pres);

	UnLock(lock_pres_record,
	"UnLocking after successfully removing for presentity");
	
        delete ret;
        return 1;
}

void DataManager::Lock(pthread_mutex_t mut,char* logMsg)
{
	debug(DEBUG_IMPP, "DataManager::Lock","LOCKING %s\n",logMsg);
	pthread_mutex_lock(&mut);
}
void DataManager::Lock(pthread_mutex_t *mut,char* logMsg)
{
	debug(DEBUG_IMPP, "DataManager::Lock","LOCKING %s\n",logMsg);
	pthread_mutex_lock(mut);
}
void DataManager::UnLock(pthread_mutex_t *mut,char* logMsg)
{
	debug(DEBUG_IMPP, "DataManager::UnLock","UNLOCKING %s\n",logMsg);
	pthread_mutex_unlock(mut);
}
void DataManager::UnLock(pthread_mutex_t mut,char* logMsg)
{
	debug(DEBUG_IMPP, "DataManager::UnLock","UNLOCKING %s\n",logMsg);
	pthread_mutex_unlock(&mut);
}
