#ifndef _DATA_MANAGER
#define _DATA_MANAGER
//#include "presds.h"
#include "augmented_node.h"
#include <pthread.h>
#include <iostream>
#include <map>
using namespace std;


//////////////////////////
//Newly added for RLS
/////////////////////////
#include "presenceagent.h"
#include "rls_presds.h"


class PresentityDS;
/**
 * Server_conf structure.
 * This structure is used to store the address of Database, the XCAP server
 * and the proxy server.
 */
typedef struct server_conf {
	char* proxyAddr;
	char* dbAddr;
	char* xcapAddr;
};
/**
 * subscription_state_t enum for subscription states.
 */
typedef enum subscription_state_t1 {
  blocked,
  pending,
  active,
  unknown
} subscription_state_t;
/**
 * InMemory Data manager.
 * Single instance of this clas in the system.
 * This class is used for in memory data store.
 * provides API to create, remove and get presentity object.
 * It also provides locking API's for presentity object.
 * There is a global lock on the in memory store.
 */
class DataManager {


	public :

	  ///////////////////////////////
	  //Added for RLS
	  ///////////////////////////////
	//Map of the subscriber id to list subscription
	static map<string, string> list_sub_record;
	static string getTheResourceList(string key);
	static void setTheResourceList(string, string);

	//For RLS presentity
	static map<string, RLSPresentityDS*> rls_pres_record;
	static void* createRLSPresentity(string);
	static RLSPresentityDS* getRLSPresentityDS(string);
	static int removeRLSPresentity(string);
	
	//For Document of the list subscriber
	static map<string, vector<string> > rls_resources_record;
	static vector<string> getResources(string);
	static void setResources(string, vector<string>);





	/// Map of presentity id to PresentityDS objects
	static map<string,PresentityDS*> pres_record;

	/// Lock on the map of presentity objects
	static pthread_mutex_t lock_pres_record;
	
	DataManager();

	~DataManager();
	
	/// @returns the instance of DataManager
	static	DataManager* getInstance();

	static void setServerConf(struct server_conf*);
	static struct server_conf*  getServerConf();
	
	/// Creates a container and associated objects for a presentity.
	static void* createPresentity(const string & presentity);

	/// gets the presentity, the calling function need to use presentity lock
	static PresentityDS* getPresentityDS(const string &  presentity);

	/// delete the presentity, the calling function need to use global lock
	static int removePresentity(const string & presentity);

	/// checks if presentity exists, internally lock is used.
	static bool ifExists(const string & presentity);

	/// Lock and Unlock take a string to log the messages
	static void Lock(pthread_mutex_t*,char*);

	static void UnLock(pthread_mutex_t*,char*);
	
	/// Lock, string to log the messages
	static void Lock(pthread_mutex_t,char*);

	/// Lock, string to log the messages
	static void UnLock(pthread_mutex_t,char*);

	static DataManager *dmInstance;
	static struct server_conf *sconf;

	private :


};
#endif
