#include "presenceagent.h"
#include "publish_base_handler.h"
#include "publish_pres_handler.h"
#include "subscribe_pres_handler.h"
#include "subscribe_base_handler.h"
#include "subscribe_winfo_handler.h"
#include "presds.h"

///////////////////////////////////////
//Added for RLS
////////////////////////////////////////
#include "subscribe_eventlist_handler.h"
#include "DataManager.h"

using namespace std;
////////////////////////////////////////

extern "C" {
#include "thread_pool.h"
}

#define DB_BASED_PUBLISH 1

MyPresenceAgent* MyPresenceAgent::end_point;
MySIP_PA *globalcall;

MyPresenceAgent::MyPresenceAgent(): SIPSessionEndpoint()
{
  MyPresenceAgent::end_point = this;
}

MyPresenceAgent* MyPresenceAgent:: GetSIPEndPoint()
{
  return end_point;
}

IptelCall *MyPresenceAgent::AllocateNewCallObject(const string callid)
{
  debug(DEBUG_IMPP, "MyPresenceAgent::AllocateNewCallObject", "creating MySIP_PA\n");
  return new MySIP_PA(*this, callid);
}

SIPDialog *MyPresenceAgent::AllocateNewDialogObject(SIPServerTransaction & transaction, SIPMessage & message)
{
  debug(DEBUG_IMPP, "MyPresenceAgent::AllocateNewDialogObject","creating new dialog obj\n");
  
  if (message.IsRequest()) {
    SIPRequestMessage req = message;
    
    switch(req.GetMethodType())
      {
      case SIPRequestMessage::SUBSCRIBE:
      case SIPRequestMessage::NOTIFY:
      case SIPRequestMessage::PUBLISH:
	{
	  debug(DEBUG_IMPP,"MyPresenceAgent::AllocateNewDialogObject",
		"creating new call obj\n");
	  SIPHeader callid_header = message.GetHeader("Call-ID");
	  string callid = callid_header.GetValue();

	  /* We use the SIPSession for this */
	  return (SIPDialog *) AllocateNewCallObject(callid);
	}
      default:
	break;
      }		
  }
  debug(DEBUG_IMPP, "MyPresenceAgent::AllocateNewDialogObject",
	"No match->SIPEndPointAllocateNewDialog\n");
  return SIPEndpoint::AllocateNewDialogObject(transaction, message);
}

bool MyPresenceAgent ::OnReceiveInitialRequest(SIPServerTransaction & transaction,  SIPRequestMessage & message)
{
  debug(DEBUG_IMPP, "MyPresenceAgent::OnReceiveInitiaRequest", "entered\n");
  return SIPEndpoint::OnReceiveInitialRequest(transaction,message);
}




MySIP_PA::MySIP_PA(MyPresenceAgent & ep, const string callid) : SIPSession(ep, callid)
{
  debug(DEBUG_IMPP, "MySIP_PA::MySIP_PA","Creating dialog object %p\n",this);
}

MySIP_PA::~MySIP_PA()
{
  debug(DEBUG_IMPP, "MySIP_PA::Destructor","Destroying dialog object %p\n",this);
}

// return true to indicate message handled by this layer
bool MySIP_PA::OnReceiveRequest(SIPRequestMessage & msg)
{
  debug(DEBUG_IMPP, "MySIP_PA::OnReceiveRequest", "entered\n");
  bool ret = true;

  ret = SIPDialog::OnReceiveRequest(msg);

  switch(msg.GetMethodType()) {
  case SIPRequestMessage::PUBLISH:
    debug(DEBUG_IMPP,"MySIP_PA::OnReceiveRequest","message type is PUBLISH\n");
    ret = OnReceivePublish(msg);
    break;

  case SIPRequestMessage::SUBSCRIBE:
    debug(DEBUG_IMPP,"MySIP_PA::OnReceiveRequest","message type is SUBSCRIBE\n");
    ret = OnReceiveSubscribe(msg);
    break;

  case SIPRequestMessage::NOTIFY: 
    debug(DEBUG_IMPP,"MySIP_PA::OnReceiveRequest","message type is NOTIFY\n");
    ret = OnReceiveNotify(msg);
    break;

  default:
    debug(DEBUG_IMPP, "MySIP_PA::OnReceiveRequest", "entered default\n");
    //SIPDialog(header parsing) -> SIPCall(invite/bye) ->SIPSession(sub/notify)
    if(ret==true){
      debug(DEBUG_IMPP, "MySIP_PA::OnReceiveRequest","message already handled by SIPDialog\n");
    }
    break;
  }
  return ret;
}

//return true to indicate message handled at this layer
bool MySIP_PA:: OnReceivePublish(SIPRequestMessage &msg)
{
  debug(DEBUG_IMPP, "MySIP_PA::OnReceivePublish", "entered\n");
  return OnPublish(msg);
}

bool MySIP_PA:: OnReceiveSubscribe(SIPRequestMessage& msg)
{
  debug(DEBUG_IMPP, "MySIP_PA::OnReceiveSubscribe", "entered\n");
  bool ret = true;

  SIPResponseMessage response;
  ret = OnSubscribe(msg,response);
  
  return ret;

}

bool MySIP_PA::OnSubscribeSuccess(const SIPResponseMessage& response)
{
  debug(DEBUG_IMPP,"MySIP_PA::OnSubscribeSuccess",
	"Your subscription to %s succeeded  \n",GetRemoteAddress().AsString().c_str());

  return false;
  
  /*
    TODO: 
    Do not delete the object.
    Useful to match with EventRegNotification:
    We may not need to store the subscription though if we rely on server to send
    notification terminated or check on next subscribe.
  */
}

bool MySIP_PA::OnSubscribeFailed(const SIPResponseMessage& response)
{
  debug(DEBUG_IMPP,"MySIP_PA::OnSubscribeSuccess",
	"Your subscription to %s failed %s \n",GetRemoteAddress().AsString().c_str(),
	response.GetReasonPhrase().c_str());

  Terminate(); // deleting the object
  return false; /* we already called terminate(), 
		   so not asking lib to delete the object */
}

bool MySIP_PA::OnSubscribeExpired()
{
  debug(DEBUG_IMPP,"MySIP_PA::OnSubscribeSuccess",
	"Your subscription to %s expired \n",GetRemoteAddress().AsString().c_str());
  return true; /* will delete the object */
}

//TODO:Test code for perf measurement: will be removed
void MySIP_PA::TrialNotify()
{
  string from="A@128.59.19.159";
  string to ="callhandler@128.59.19.159";
  string to_contact="callhandler@128.59.19.159:6786";
  string body =" This is test notify";
  string notify_state ="active";
  string event_type ="presence";
  
  debug(DEBUG_IMPP, "SendNotification::SendNotify", "Notify body %s\n",
	body.empty() ? "NULL" : body.c_str());

  //MySIP_PA *call = new MySIP_PA(*(MyPresenceAgent::GetSIPEndPoint()));
  MySIP_PA *call = this;//globalcall;
  
  string remote_addr =to_contact;
  IptelAddress remote_uri = remote_addr;
  remote_uri.SetScheme("sip");
  call->SetRemoteURI(remote_uri);
  
  string remote = to;
  IptelAddress remoteurl    = remote;
  remoteurl.SetScheme("sip");
  call->SetRemoteAddress(remoteurl);
  
  string src = from;
  IptelAddress srcurl    = src;
  srcurl.SetScheme("sip");
  call->SetSelfAddress(srcurl);
  
  //call->SetPreferredTransportType(SOCK_TLS);
  //call->SetPreferredTransportType(SOCK_DGRAM);
  //call->SetPreferredTransportType(SOCK_STREAM);
  
  call->TestNotify(event_type,body);
  
  //TODO:: Terminate OnNotifySuccess : changes based on this
  debug(DEBUG_IMPP, "SendNotification::SendNotify", "Send Notify done\n");
  
}

bool MySIP_PA::Notify(const string& event,const string & body1,
		const string& content_type,const string & state)
{
  SIPRequestMessage msg = BuildRequest("NOTIFY");

  SIPMessageBody body;
  body.SetContentType(content_type);
  body.SetContent(body1);
  msg.SetBody(body);

  SIPHeader event_header("Event", event);
  msg.SetHeader(event_header);

  SIPHeader subscription_state_header("Subscription-State",state);
  msg.SetHeader(subscription_state_header);

  SendRequest(msg);
  return true;
}


//////////////////////////////////
//Added for RLS
//////////////////////////////////
bool MySIP_PA::RLSNotify(const string& event,const string & body1,
		const string& content_type,const string & state)
{
  SIPRequestMessage msg = BuildRequest("NOTIFY");

  SIPMessageBody body;
  body.SetContentType(content_type);
  body.SetContent(body1);
  msg.SetBody(body);

  SIPHeader event_header("Event", event);
  msg.SetHeader(event_header);

  SIPHeader subscription_state_header("Subscription-State",state);
  msg.SetHeader(subscription_state_header);

  //The 'Require' header field for RLS
  SIPHeader required_header("Require", "eventlist");
  msg.SetHeader(required_header);
  
  SendRequest(msg);
  return true;
}


// Debugging code added to see NOTIFY problem
#if 1
struct send_notify_arg {
  string pres;
  string body;
  MySIP_PA *call;
  long count;
};

void *send_notify(void *arg)
{
	send_notify_arg *a = (send_notify_arg *)arg;

        SIPMessageBody body;
	body.SetContentType("application/pidf+xml");
	body.SetContent(a->body);
        lock_dialog(*a->call);//pthread_mutex_lock(&mutex);
        a->call->SIPSession::Notify(a->pres,body);
        unlock_dialog(*a->call);//pthread_mutex_unlock(&mutex);
	
	  //a->call->TestNotify(a->pres,a->body);
	  --(a->count);
	if (a->count % 100 == 0) {
		printf("remaining %d more\n", a->count);
	}
	else if (a->count > 999900) {
		printf("current %d\n", a->count);
  	}

	if (a->count > 0) {
		if (pool_schedule(send_notify, arg) != 0) {
			printf("queue full, so calling inline\n");
			send_notify(arg);
    		}
	}

}
#endif

//TODO: to be replaced by other method
void MySIP_PA::TestNotify(const string & pres,const string & body1)
{
  debug(DEBUG_IMPP, "MySIP_PA::TestNotify", "entered\n");
	
  //const string pres ="presence";
  SIPMessageBody body;
  body.SetContentType("application/pidf+xml");
  body.SetContent(body1);
	
  //SetPreferredTransportType(SOCK_STREAM);
  pthread_mutex_lock(&mutex);//Dialog LOCKED
  SIPSession::Notify(pres, body);
  pthread_mutex_unlock(&mutex);//UNLOCKED after NOTFIY
  debug(DEBUG_IMPP, "MySIP_PA::TestNotify", "done\n");

// Debugging code
#if 0
  send_notify_arg *arg = new send_notify_arg;
  arg->pres = pres;
  arg->body = body1;
  arg->call = this;
  arg->count = 1000000;
  pool_schedule(send_notify, arg);
#endif
}

///////////////////////////
//Added for RLS
///////////////////////////
void MySIP_PA::RLSSubscribe()
{
  SIPRequestMessage msg = BuildRequest("SUBSCRIBE");

  msg.SetHeader(SIPHeader("Event", "presence"));
  msg.SetHeader(SIPHeader("Supported", "eventlist"));


  //Get the URI to find RLSPresentity
  SIPHeader header_to = msg.GetHeader("To");
  string to_userdomain = header_to.AsString(); 
  IptelAddress to_useraddr = to_userdomain;
  to_userdomain = to_useraddr.GetUserHost();

  DataManager *dm = DataManager::getInstance();
  RLSPresentityDS* rpds = dm->getRLSPresentityDS(to_userdomain);

  //Check 'unsubscription' of the presentity, 
  //if true, set 'Expriese' header field to 0, otherwise, set default value 
  if ((rpds->getUnsubscription()) == true){
    msg.SetHeader(SIPHeader("Expires", "0"));
  } else {
    msg.SetHeader(SIPHeader("Expires", "3600"));
  }

  //TODO: Multiple 'Accepted' Header fields should be added, 
  //      but library doesn't allow it 
  msg.SetHeader(SIPHeader("Accepted", "application/pidf+xml"));
  //msg.SetHeader(SIPHeader("Accepted", "application/rlmi+xml"));
  //msg.SetHeader(SIPHeader("Accepted", "multipart/related"));

  subscribe_state = SUBSCRIBE_SENT;
  SendRequest(msg);  

}




void MySIP_PA::Publish(const string& event, const SIPMessageBody& body)
{
	SIPRequestMessage msg = BuildRequest("PUBLISH");
	msg.SetBody(body);
	SIPHeader event_header("Event", event);
	msg.SetHeader(event_header);
	SIPHeader expires_header("Expires", "3600");
	msg.SetHeader(expires_header);
	SendRequest(msg);  

}


// This method takes care of segregation of logic, from SIPSession 
bool MySIP_PA :: OnReceiveNotify(SIPRequestMessage &msg)
{
  debug(DEBUG_IMPP, "MySIP_PA::OnReceiveNotify", "entered\n");
  bool ret = true;

  SendResponse(200, "OK");
  OnNotify(msg);
  return ret;
}


///////////////////////////
//Added for RLS
///////////////////////////
void MySIP_PA::OnNotify(const SIPRequestMessage& request)
{
  //Receive the NOTIFY message from the resource the local RLS sends a SUBSRIBE to
  string content_type = "";
  string state = "";
  string strBody = "";

  if (request.HasBody()) {
    SIPMessageBody body = request.GetBody();
    debug(DEBUG_IMPP,"ListPresentity::OnNotify",
	  "Message contains the information. Body Message is %s \n", 
	  body.GetContent().c_str());

    state = "active";
    strBody = body.GetContent();
  } 
  //The External Server doesn't have enough information 
  //to reject or accpet the subscription
  else {  
    debug(DEBUG_IMPP,"ListPresentity::OnNotify",
	  "External Server has Not enough information. \n");
    
    state = "pending";
    strBody = "";
  }


  //Get the URI to find RLSPresentity
  SIPHeader header_from = request.GetHeader("From");
  string from_user_domain = header_from.AsString(); 
  IptelAddress from_user_addr = from_user_domain;
  from_user_domain = from_user_addr.GetUserHost();

  //Get the RLSPresentity to forward NOTIFY message. (Get the incoming dialog)
  DataManager *dm = DataManager::getInstance();
  RLSPresentityDS* rpds = dm->getRLSPresentityDS(from_user_domain);
  if(rpds == NULL) {      
    error("MySIP_PA::OnNotify", "Can't find the presentity %s\n", 
	    from_user_domain.c_str());
    return;
  }
  
  //Send NOTIFY on that dialog(imcoming call)
  (rpds->getIncomingCall())->RLSNotify("presence", strBody, 
				       "application/pidf+xml", state); 

  //Check 'unsubscription' of the presentity, 
  //if true, delete presentity and subscription
  if ((rpds->getUnsubscription()) == true){
    //Remove all resources the presentity might hold
    Terminate();

    int ret = dm->removeRLSPresentity(from_user_domain);
    if(ret == -1){
      error("MySIP_PA::OnNotify", "Can't remove the presentity %s\n", 
	    from_user_domain.c_str());
    } else {
      debug(DEBUG_IMPP,"MySIP_PA::OnNotify", 
	    "Successfully removed rls presentity: %s\n", 
	    from_user_domain.c_str());
    }
  }
  debug(DEBUG_IMPP,"MySIP_PA::OnNotify", "Done with 'OnNotify'\n");
}


bool MySIP_PA::OnNotifySuccess(const SIPResponseMessage& response)
{
	debug(DEBUG_IMPP, "MySIP_PA::OnNotifySuccess", "Your notify  to %s completed\n",
		GetRemoteAddress().AsString().c_str());
        //return true;// for use with different dialogs
	return false;
	 /* Do not delete the object 
	    As u want to use the object to send NOTIFY when
	    PUBLISH is received using the same Dialog object
	 */
}

bool MySIP_PA::OnNotifyFailed(const SIPResponseMessage& response)
{
	error("MYSIP_PA::OnNotifyFailed", "Notify to %s failed, because  %s \n",
	GetRemoteAddress().AsString().c_str(),response.GetReasonPhrase().c_str());
	
        //return true;/// for use with different dialogs
	///TODO:XXX:Free Subscriber Object here

	//Terminate(); /// TODO : May need to free the dialog object from SubscriberDS
	return false; /* will delete the object */
}

bool MySIP_PA::OnFailureResponse(SIPResponseMessage & msg) 
{
	int status = msg.GetStatusCode();

	debug(DEBUG_IMPP,"MySIP_PA::OnFailureResponse", "Response status %d  %s  \n",status,
		msg.GetReasonPhrase().c_str());

	if (status == 401 || status == 407) {
		SIPHeader h = msg.GetHeader("WWW-Authenticate");
	}
	else if (msg.Is3xxResponse()) {
	    SIPContactHeader h;
	}

	return SIPSession::OnFailureResponse(msg);
}

/* PresenceHandling Algorithm

Create the handler  and Start the handler machinery
    Encapsulate this code with 
	- createHandler( ) as a central place to be called
	- which checks 
		- Message method type
		- Event package type
		- Presentity - for composition info
	- creates a PresenceHandler type object and invokes
		- execute()
	 	- invoking execute() can be replaced by putting into queue.
		- to see comparision of two approaches.
		- queuign may be needed for 
			- offline processing to prevent dropping of SIP messages to
			processing of messages
			- persisting messages in DB to process later or getting 
			processed by remote machines - load balanced.
			- queue depth to be criteria compared to number of threads for
			the message to be dropped.

	- Presencehandler will be passed to a genric machinery which processes by calling
		ph->execute()

	Look for complete impl in RFC 3903
*/

// returns true to indicate message handled at this layer
bool MySIP_PA::OnPublish(const SIPRequestMessage &request)
{
  //Test Code for Performance without presence processing.
  //SendResponse(200,"OK");
  //TrialNotify();
  //TrialNotify();
  //Terminate();
  //return true;
	
#ifdef DB_BASED_PUBLISH
/*
  PresenceHandler *ph = (PresenceHandler*)new PublishHandlerDB((SIPRequestMessage)request,this);
  ph->execute();
  debug(DEBUG_IMPP, "MyPresenceAgent::OnPublish", "Publish Handler Functions executed \n");
  delete ph;
  Terminate();
  return true;
*/
#endif

  PresenceHandler *ph =NULL;
  int event_type = getEventType(request);
  if(event_type==EVENT_PRESENCE){
    ph = (PresenceHandler*)new PublishPresHandler((SIPRequestMessage)request,this);
  } else if(event_type==EVENT_E911STATUS){
    ph = (PresenceHandler*) new PublishBaseHandler((SIPRequestMessage)request,this);
  }else {
    SendResponse(489,"Bad Event Header");
    Terminate();
    return true;
  }

  ph->execute();
  debug(DEBUG_IMPP, "MyPresenceAgent::OnPublish", "Publish Handler Functions executed \n");

  delete ph;
  Terminate();

  return true;
}

int MySIP_PA:: getEventType(const SIPRequestMessage & request)
{
  int event_type = EVENT_UNKNOWN;
  SIPHeader header_event = request.GetHeader("Event");
  string event_name = header_event.GetValue();
  
  if(strcasecmp((char*)event_name.c_str(),(char*)"presence")==0){
    event_type = EVENT_PRESENCE; 
  }else if(strcasecmp((char*)event_name.c_str(),(char*)"lookup")==0){
    event_type = EVENT_LOOKUP;
  }else if(strcasecmp((char*)event_name.c_str(),(char*)"E911status")==0){
    event_type = EVENT_E911STATUS;
  }else if(strcasecmp((char*)event_name.c_str(),(char*)"presence.winfo")==0){
    event_type = EVENT_PRESENCE_WINFO;

  }else {
    event_type =EVENT_UNKNOWN;
    debug(DEBUG_IMPP,"MySIP_PA::getEventType",
	  "unknown event type  %s \n", event_name.c_str()) ;
  }
  debug(DEBUG_IMPP,"MySIP_PA::getEventType",
	"event type  %s set is %d \n", event_name.c_str(),event_type) ;
  return event_type;
}

////////////////////////////////
// Added for RLS. Local Function
////////////////////////////////
bool ValidateResourceList(string ID, string listID){
  bool ret = false;

  DataManager* dm = DataManager::getInstance();
  string listIDInDB = dm->getTheResourceList(ID);

  if ( listIDInDB == listID){
    ret = true;
  } 

  return ret;  
}


///////////////////////////
//Edited for RLS
///////////////////////////
bool MySIP_PA::OnSubscribe(const SIPRequestMessage& request, 
			   SIPResponseMessage& response)
{
  PresenceHandler *ph = NULL;
  int event_type = getEventType(request);

  /// TODO: Get the right object and implement a clone interface
  if(event_type==EVENT_PRESENCE) {

    /////////////////////////////////
    // Added for RLS
    /////////////////////////////////
    SIPHeader header_from = request.GetHeader("From");
    string from_user = header_from.AsString();
    IptelAddress from_user_addr = from_user;
    from_user = from_user_addr.GetUser();

    SIPHeader header_to = request.GetHeader("To");
    string to_user = header_to.AsString();
    IptelAddress to_user_addr = to_user;
    to_user = to_user_addr.GetUser();


    //Check if the ID from 'To' Header field matches the list name 
    //If so, call 'SubscribeEventlistHandler', otherwise call 'SubscribePresHandler'
    bool validity = ValidateResourceList(from_user, to_user);
    if (validity) { 
      //TODO: Need to change it into the code that commented out, 
      //      since, for now, client doesn't support 'Supported' header field
      //SIPHeader header_supported = request.GetHeader("Supported");
      //string strSupported = header_supported.GetValue();
      string strSupported = "eventlist";
      if (strSupported == "eventlist") {
	ph = (PresenceHandler*) 
	  new SubscribeEventlistHandler((SIPRequestMessage)request,this);
      }	else {
	SendResponse(421,"Extenstion Required");
	return false;
      }
    } else {
      ph = (PresenceHandler*) 
	new SubscribePresHandler((SIPRequestMessage)request,this);    
    }

  } else if (event_type==EVENT_PRESENCE_WINFO){
    ph = (PresenceHandler*) 
      new SubscribeWinfoHandler((SIPRequestMessage)request,this);
  } else if(event_type==EVENT_E911STATUS){
    ph = (PresenceHandler*) 
      new SubscribeBaseHandler((SIPRequestMessage)request,this);
    
  } else {
    SendResponse(489,"Bad Event Header");
    return false;
  }
  
  ph->setDialog(this);
  ph->execute();

  debug(DEBUG_IMPP, "MyPresenceAgent::OnSubscribe", 
	"Subscriber Handler Functions executed \n");
  delete ph;
  
  return true; 
}
