#ifndef _PRESENCEAGENT
#define _PRESENCEAGENT
extern "C" {
#include <pthread.h>
#include "timespec.h"
#include "error.h"
#include "strdupn.h"

extern char *optarg;
}

#include <stdio.h>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <list>

#include <mysql.h>
#include "iptelcall.h"
#include "sipmessage.h"
#include "iptelsession.h"
#include "sipcallobj.h"
#include "sipregobj.h"
#include "sipendpoint.h"
#include "siplibrary.h"
#include "sipsession.h"
#include "sipcpim.h"


//#include "radio.h"   
using namespace std;


//this is global
class SIPEndpoint;
/**
 * MyPresenceAgent : A SIPSessionEndPoint class.
 */

class MyPresenceAgent : public SIPSessionEndpoint
{
 public:
  /**
   * EndPoint Instance.
   * EndPoint Constructor: To create a new SIP Stack
   */
  MyPresenceAgent();
  
  /**
   * Returns the end point instance.
   * Returns the end point instance which is needed to create Presence Dialog
   */
  static MyPresenceAgent* GetSIPEndPoint();
  
  /**
   *	Allocates new call object.
   */
  IptelCall *AllocateNewCallObject(const std::string callid);
  
  /**
   *	Allocates new dialog object.
   */
  SIPDialog *AllocateNewDialogObject (SIPServerTransaction & transaction, SIPMessage & message);
  /**
   *	Overridden -> receive initial request.
   */
  bool OnReceiveInitialRequest (SIPServerTransaction & transaction, SIPRequestMessage & message);
  
  static MyPresenceAgent* end_point; 
  friend class MySIP_PA;
  
 protected:
  int  call_num;
};
/**
 *  MySIP_PA : A sip session class, represents a dialog.
 */
class MySIP_PA : public SIPSession
{
 public:
  /**
   *  MySIP_PA : A sip session class, represents a dialog.
   *  The Session object is created when a message is received or need to be sent.
   *  The call back function are invoked by the stack i.e end point object when message is received.
   *  It provides functions to be called for sending response and sending messages by creating a new session.
   */
  MySIP_PA(MyPresenceAgent & ep, const std::string callid = "");
  
  ~MySIP_PA();
  
  
  
  /**
   * This function is called on a session object when request message is recieved.
   */
  bool OnReceiveRequest(SIPRequestMessage & msg);
  
  /**
   *  Call back method when failure response is received.
   *  If a derived class wants to do additional handling on receiving failure response,
   *  thi function can be modified or overridden.
   */
  
  bool   OnFailureResponse(SIPResponseMessage & msg);
  
  /**
   * This function is called when SUBSCRIBE message is received.
   */
  bool   OnReceiveSubscribe(SIPRequestMessage& msg);
  /**
   * This function is called to invoke the correct SubscribeHandler.
   * @see SubscribeHandler  -- presence event package
   * @see BaseSubscribeHandler -- Default package
   * Unknown event packages are handled using a default package implementation.
   */
  bool   OnSubscribe(const  SIPRequestMessage& request,
                     SIPResponseMessage& response);
  /**
   * This function is called when a successfull response for SUBSCRIBE message is received.
   * This will be useful when the Presence Server subscribes to the registrar for Event Registration
   * package.
   */
  bool   OnSubscribeSuccess(const SIPResponseMessage& response);
  /**
   * This function is called when SUBSCRIBE fails.
   */
  bool   OnSubscribeFailed(const SIPResponseMessage& response);
  
  bool   OnSubscribeExpired();
  /**
   * parses the request and gives the event type
   * used to get the correct subscribe/publish handler type.
   */
  int getEventType(const SIPRequestMessage & request);
  
  /**
   * This function is called when NOTIFY message is received.
   */
  bool OnReceiveNotify(SIPRequestMessage &msg);
  
  /**
   * This function is called when  successfull response for NOTIFY message is received.
   */
  bool   OnNotifySuccess(const SIPResponseMessage& response);
  
  /**
   * This function is called when  failure response for NOTIFY message is received.
   */
  bool   OnNotifyFailed(const SIPResponseMessage& response);
  
  /**
   *	This function invokes NotifyHandler.
   *   The Notify handler needs to maintain dialog state of SUBSCRIBE's it sent.
   */
  void   OnNotify(const SIPRequestMessage& request);
  
  /**
   * This function can be used to send PUBLISH
   * @see PublishHandler  -- presence event package
   * @see BasePublishHandler -- Default package
   * Unknown event packages are handled using a default implementation
	    * Which means we never send 415 bad event for authneitcated PUBLISH.
	    */	
  void   Publish(const std::string& event, const SIPMessageBody& body);




  //////////////////////////
  //Added for RLS
  /////////////////////////
  void RLSSubscribe();
  bool RLSNotify(const std::string& event,const std::string& body,
	      const std::string& content_type,const std::string& substate);



  
  
  /**
   * Function to send notify.
   * Takes event type, message body, content-type header,
   * Subscription-state header.
   */
  bool Notify(const std::string& event,const std::string& body,
	      const std::string& content_type,const std::string& substate);


  
  /**
   *  This function is invoked to process received PUBLISH request.
   */
  bool   OnReceivePublish(SIPRequestMessage &msg);
  
  /**
   *  This function invokes the PUBLISH handler.
   *  The handler takes care of different event packages.
   */
  bool   OnPublish(const SIPRequestMessage &request);
  
  void TestNotify(const string &, const string &);
  void TrialNotify();
  
  friend int main(int argc, char *argv[]);
  
  
 protected:
};

//MySIP_PA *globalcall;

/*static std::string global(const std::string & key)
  {
  std::map<std::string,std::string>::iterator gi;
  gi = globals.find(key);
  if (gi == globals.end()) {
  return "";
  }
  else {
  return (*gi).second;
  }
  }*/
#endif
