#include "send_eventlist_notification.h"
#include "p_handler.h"
#include "DataManager.h"

SendEventlistNotification::SendEventlistNotification()
{

}
SendEventlistNotification::SendEventlistNotification(PresenceHandler *ph, 
						     PresentityDS* pds)
{
  debug(DEBUG_IMPP, "SendEventlistNotification::", "Constructor entered\n");
  pHandler = ph;
  preDS = pds;
}

SendEventlistNotification::~SendEventlistNotification()
{
}

void SendEventlistNotification::create_and_send_notify()
{	
  debug(DEBUG_IMPP, "SendEventlistNotification::create_and_send_notify", 
	"Just entered\n");

  string state = "";	
  string body = "";
  string pres = "presence";    
 
  SubscriberDS* sds = preDS->getSubscriberDS(EVENT_PRESENCE, 
					     pHandler->get_from_userdomain());
  if(sds==NULL){
    error("SendEventlistNotification::create_and_send_notify",
	  "Could not find subscriber record in the memory DB \n");
    return;
  }
	
  debug(DEBUG_IMPP, "SendEventlistNotification::create_and_send_notify",
	"sending notifications for presentity %s\n", 
	preDS->getPresentityName().c_str());

  
  if(pHandler->getSubscriptionState() == pending){
    debug(DEBUG_IMPP, "SendEventlistNotification::create_and_send_notify", 
	  "subscribe -> notification state is pending\n");

    state = "pending";
    
  } else if(pHandler->getSubscriptionState() == active){
    debug(DEBUG_IMPP, "SendEventlistNotification::create_and_send_notify", 
	  "subscribe -> notification state is active\n");

    body = prepareActiveNotify(preDS, sds);
    state = "active";

  }else {
    error("SendEventlistNotification::create_and_send_notify",
	  "Don't know the subscription state\n");
    return;
  }
      
  //Call 'RLSNotify' method on the Dialog created for the subscriber
  (pHandler->getDialog())->RLSNotify(pres, body, "application/pidf+xml", state);
 
  debug(DEBUG_IMPP, "SendEventlistNotification::create_and_send_notify",
	"Done with creating and sending notify\n" );
}

string SendEventlistNotification:: prepareActiveNotify(PresentityDS* pds, 
						       SubscriberDS *sds)
{
  int event = pHandler->getEvent();
  char* body = NULL;
  string strBody ="";

  struct server_conf *conf =  DataManager::getServerConf();
  XMLParser * xparser = pHandler->getXMLParser();


  pds->LockData(event);

  AugmentedDOMNode *adn = (AugmentedDOMNode*) pds->getPresenceData(event);
  if(adn==NULL){
    error("SendEventlistNotification::prepareActiveNotify", 
	  "ADN was NULL, NOTIFY not sent \n" );
    pds->UnLockData(event);
    return "";
  }
  
  DOMNode *presenceState = adn->getPresenceBody(); 
  if(presenceState==NULL){
    error("SendEventlistNotification::prepareActiveNotify",
	  "DOM body NULL, NOTIFY not sent \n" );
    pds->UnLockData(event);
    return "";
  }

  body = xparser->serializeDOM(presenceState);
  if(body!=NULL){
    strBody = body;
    XMLString::release(&body);
  } 
  
  debug(DEBUG_IMPP, "SendEventlistNotification::prepareActiveNotify", 
	"Server %s\n", conf->proxyAddr);
  
  pds->UnLockData(event);
  
  return strBody;
}

int SendEventlistNotification:: apply()
{
  debug(DEBUG_IMPP, "SendEventlistNotification::apply", "entered\n"); 
  create_and_send_notify();
  debug(DEBUG_IMPP, "SendEventlistNotification::apply", "Done\n");
  
  return 1;
}

