#include "query_constructor.h"
#include "DBManager.h"

#include "send_eventlist_notification.h"
#include "unsubscription.h"
#include "subscribe_eventlist_handler.h"
#include "xmlparser.h"


SubscribeEventlistHandler::SubscribeEventlistHandler()
{
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::Constructor",
	"This is the default constructor");
}

SubscribeEventlistHandler::SubscribeEventlistHandler(void* mySIP_pa)
  :SubscribePresHandler(mySIP_pa)
{
  
}

SubscribeEventlistHandler::SubscribeEventlistHandler(const SIPRequestMessage & request)
{
  
}


SubscribeEventlistHandler::SubscribeEventlistHandler(const SIPRequestMessage & request,
			     void* mySIP_pa):SubscribePresHandler(mySIP_pa)
{
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::constructor", "Just entered\n");

  //Get host name of the server
  char* hostname;
  gethostname(hostname, 50);
  string server_domain = hostname;

  //Get the 'Expires' header field 
  SIPHeader header_expires = request.GetHeader("Expires");
  string expires = header_expires.GetValue();

  setMethod(METHOD_SUBSCRIBE);  
  parseRequest(request);
  
  //First of all, need to check validation and authorization of the message
  //It is the same procedure as 'the SubscribePresHandler' does
  int validation = validateSubscription(request);  
  if(validation == -1){
    error("SubscribeEventlistHandler::Constructor",
	  "Invalid headers or body in SUBSCRIBE\n");
    SendResponse(400, "Bad Request Error");
    return;
  } 
  
  //It will send '200 OK' message if the request gets an authorization
  int authorization = authorizeSubscription(request);
  if(authorization == -1){
    error("SubscribeEventlistHandler::Constructor",
	  "Failed to authorize Subscription\n");
    SendResponse(500, "Internal Server Error");
    return;
  }


  //If this is first come, then read XML file, parse it, get the list, 
  // and save it into the DataManger. 
  // otherwise, just get the list from the DataManager. 
  //Assumed 'to_user' is same as the XML file name
  string to_user = get_to_user();
  vector<string> resources = dm->getResources(to_user);
  if (resources.size() == 0) {
    resources = getResources(to_user + ".xml");
    dm->setResources(to_user, resources);
  }
  
  vector<string>::const_iterator citer;
  for(citer=resources.begin(); citer!=resources.end(); citer++){
    string to_userdomain = *citer;
    
    //Extract only its domain from the 'to_userdomain'.
    int pos = to_userdomain.find("@", 0);
    string to_domain = 
      (to_userdomain).substr(pos + 1, (to_userdomain).size() - (pos + 1));
    
    
    //The case of User is not in this domain
    if (to_domain != server_domain){
      string to_contact = to_userdomain;
      
      MySIP_PA* outgoingCall = NULL;
      RLSPresentityDS* rpds = NULL;
      //In case of the User does not exist in Memory 
      if (( rpds = checkRLSUserInMemory(to_userdomain)) == NULL) {
	//Create a presentity and a dialog for the user
	rpds = (RLSPresentityDS*) dm->createRLSPresentity(to_userdomain);
	outgoingCall = new MySIP_PA(*(MyPresenceAgent::GetSIPEndPoint()));
	
      } 
      //In case of the User exists in Memory
      else {
	outgoingCall = rpds->getOutgoingCall();
      }
      
      //Create subscription for the RLSPresentity
      rpds->setIncomingCall(getDialog());
      rpds->setOutgoingCall(outgoingCall);

      //In the case of Unsubscription, 
      // set flag for unsubscription and remove presentity and subscription 
      // after receiving notification
      if (expires == "0"){
	rpds->setUnsubscription(true);
      }
      
      //Send a SUBSCRIBE request 
      SendSubscribe(outgoingCall, to_contact, get_from_userdomain(), to_userdomain);
    }
    
    //The case of User is in this domain
    else {
      PresentityDS *pds = checkUserInMemory(to_userdomain);
      if(pds == NULL){
	
	bool ret = checkUserInDB(to_userdomain);       
	if(ret == false){
	  //Don't Terminate the Handler. Just send a '404' Response
	  SendResponse(404,"User Not Found");
	}else {
	  debug(DEBUG_IMPP,"SubscribeEventlistHandler::constructor",
		"User found in DB, creating a presentity in the List\n");
	  
	  //This is the presentity for to_user
	  pds =(PresentityDS*) dm->createPresentity(to_userdomain);
	}
      }
      
      //Create subscription for to_user to know the who's subscriber
      setPresentityDS(pds);    
      int subscription = createSubscription(request, to_userdomain);
      if(subscription < 1){
	error("SubscribeEventlistHandler::Constructor",
	      "Server Error in creating subscription\n");
	SendResponse(500, "Internal Server Error");
	return ;
      }
      
      //Notification should be scheduled. 
      PresenceFunction *pc = NULL;  
      pc = (PresenceFunction*) new SendEventlistNotification(this, pds);
      debug(DEBUG_IMPP,"SubscribeEventlistHandler::constructor",
	    "SendEventlistNotification scheduled\n");
      PresenceHandler::functionList.push_back(pc);


      //In the case of Unsubscription, the unsubscription should be scheduled
      if (expires == "0"){
	pc = (PresenceFunction*) new UnSubscription(this, to_userdomain);
	debug(DEBUG_IMPP,"SubscribeEventlistHandler::constructor",
	      "Unsubscription scheduled\n");
	PresenceHandler::functionList.push_back(pc);
      }

    } // End of Else      
  }// End of FOR
  
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::constructor","DONE\n");
}

//Returns resource URIs in the list
vector<string> SubscribeEventlistHandler::getResources(string fileName)
{
  debug(DEBUG_IMPP, "SubscribeEventlistHandler::getResources",
	"Just entered\n");

  vector<string> resources;

  XMLParser * xparser = getXMLParser();
  char* cont = xparser->read_file(fileName.c_str());

  //TODO:Need to make sure the XML file ends with </list>
  string doc = cont;
  doc = doc.substr(0, doc.find("</list>", 0) + 7); 
  
  //Find root node first 
  DOMNode* root_node = xparser->parseXML(doc.c_str());

  DOMNode* aNode = xparser->get_next_node(root_node->getFirstChild(),"resource");
  if (aNode == NULL){
    debug(DEBUG_IMPP, "SubscribeEventlistHandler::getResources",
	  "No resources in the list. Returning empty list\n");
    return resources;
  }
  
  string user_domain = xparser->getAttributeValue(aNode, "uri");
  int pos = user_domain.find("sip:", 0);
  user_domain = (user_domain).substr(pos + 4, (user_domain).size() - (pos + 4));
  resources.push_back(user_domain);
  
  DOMNode* tempNode = xparser->get_next_node(aNode->getNextSibling(),"resource");
  while (tempNode != NULL) {
    user_domain = xparser->getAttributeValue(tempNode, "uri");
    pos = user_domain.find("sip:", 0);
    user_domain = (user_domain).substr(pos + 4, (user_domain).size()-(pos + 4));
    resources.push_back(user_domain);

    tempNode = xparser->get_next_node(tempNode->getNextSibling(),"resource");
  }

  XMLString::release(&cont);
  return resources;

  debug(DEBUG_IMPP, "SubscribeEventlistHandler::getResources",
	"Done successfully\n");

}


void SubscribeEventlistHandler::SendSubscribe(MySIP_PA* call, string to_contact,
					      string from_userdomain, 
					      string to_userdomain)
{
  debug(DEBUG_IMPP, "SubscribeEventlistHandler::create_and_send_subscribe",
	"Just entered\n");
  
  IptelAddress remote_uri = to_contact;
  remote_uri.SetScheme("sip");
  call->SetRemoteURI(remote_uri);
  
  //Set self address for SUBSCRIBE message
  IptelAddress srcurl = from_userdomain;
  srcurl.SetScheme("sip");
  call->SetSelfAddress(srcurl);

  //Set remote address for SUBSCRIBE message
  IptelAddress remoteurl = to_userdomain;
  remoteurl.SetScheme("sip");
  call->SetRemoteAddress(remoteurl);

  //Send SUBSCRIBE for the call via 'RLSSubscribe' method on the Dialog
  call->RLSSubscribe();

  debug(DEBUG_IMPP, "SubscribeEventlistHandler::RLSSubscribe",
 	"Done with creating and sending a SUBSCRIBE request\n" );
}


RLSPresentityDS* SubscribeEventlistHandler::checkRLSUserInMemory(string to_userdomain)
{
  DataManager *dm = DataManager::getInstance();
  RLSPresentityDS* rpds = NULL;
  
  if((rpds = dm->getRLSPresentityDS(to_userdomain)) == NULL) {
    debug(DEBUG_IMPP,"SubscribeEventlistHandler::checkRLSUserInMemory",
	  "RLSPresentity does not exist in memory");
  }
  return rpds;
  
}


PresentityDS* SubscribeEventlistHandler::checkUserInMemory(string to_userdomain)
{
  DataManager *dm = DataManager::getInstance();
  PresentityDS *pds = NULL;

  if((pds = dm->getPresentityDS(to_userdomain)) == NULL) {
    debug(DEBUG_IMPP,"SubscribeEventlistHandler::checkUserInMemory(str)",
	  "Presentity does not exist in memory");
  }
  return pds;  
}


bool SubscribeEventlistHandler::checkUserInDB(string to_userdomain)
{
  bool ret = false;
  DBManager *dbm = DBManager::getInstance();
  db_conn_obj *dbc = dbm->getConnectionObj();
  QueryConstructor *qc = new QueryConstructor();
  char* query = (char*) malloc(512);
  memset(query, '\0', 512);
  qc->check_user_in_db(query, 512, (char*) to_userdomain.c_str());
  debug(DEBUG_IMPP, "SubscribeEventlistHandler::checkUserInDB(str)",
	"query to check user in db %s\n", query);

  MYSQL_RES *result = dbm->read_query(dbc, query);
  if(result!= NULL){
    if(mysql_num_fields(result) > 0){
      mysql_free_result(result);
      ret = true;
    }
  }
  free(query);
  return ret;
}


int 
SubscribeEventlistHandler::createSubscription(const SIPRequestMessage & request, 
					      string to_userdomain)
{
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscription","Entered\n") ;
  
  int ret = 1;
  bool first_time = false;
  
  PresentityDS *pds = getPresentityDS();
  SubscriberDS *sds = pds->getSubscriberDS(getEvent(), get_from_userdomain());

  // This is first subscription
  if(sds == NULL){
    sds = new SubscriberDS(get_from_userdomain());

    sds->setDialog(getDialog());
    sds->setDomain(get_from_domain());

    //Set randomly chosen ID 
    char* rn = (char*) malloc(4);
    memset(rn, '\0', 4);
    sprintf(rn, "%d", (rand()%100));
    string random_num = rn;

    //Extract the 'to_user' from 'to_userdomain'
    string to_user = to_userdomain.substr(0, to_userdomain.find("@", 0));
    sds->setId(to_user + random_num);
    free(rn);
    
    //Set Contact URI
    sds->set_subsriber_contact(getEvent(), get_contact_addr());
    first_time = true;
    
    sds->setSubscriptionState(unknown);
    
    //Store the SubscriberDS into the resource(PresentityDS) 
    // the SubscriberDS is interested in
    pds->store(getEvent(), get_from_userdomain(), sds);
    debug(DEBUG_IMPP,"SubscribeEventlistHandler:createSubscription",
 	  "Subscriber %s created and added for the presentity %s \n",
 	  (get_from_userdomain()).c_str(), to_user.c_str()) ;
  }else{
    //TODO : update subscription dialog information
    sds->set_subsriber_contact(getEvent(), get_contact_addr());
    debug(DEBUG_IMPP,"SubscribeEventlistHandler:createSubscription",
 	  "subscriber %s information updated for presentity %s\n",
 	  (get_from_userdomain()).c_str(), to_user.c_str());
  }
  setSubscriberDS(sds);
  
  
  AugmentedDOMNode *adn = (AugmentedDOMNode*)sds->getData(getEvent());
  //For first fime 
  if(adn==NULL){
    ret = createSubscriptionDB(request, to_userdomain);
  }
  else {
    // TODO: Update expiration info and dialog info
  }
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscription","Done \n");
  return ret;
}


int 
SubscribeEventlistHandler::createSubscriptionDB(const SIPRequestMessage& request,
 						string to_userdomain)
{
  int ret = -1;

  debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB",
	"New subscription by %s\n",(get_from_user()).c_str());
  SIPHeader header_from = request.GetHeader("From");
  SIPHeader header_contact = request.GetHeader("Contact");
  SIPHeader header_event = request.GetHeader("Event");
  SIPHeader header_expires = request.GetHeader("Expires");
  SIPHeader header_callid = request.GetHeader("Call-ID");
	
  string watcher = header_from.AsString(); 
  string contact  = header_contact.GetValue(); 
  string event_type = header_event.GetValue();
  string expires = header_expires.GetValue();
  string callid = header_expires.GetValue();
	
  IptelAddress ipaddr = watcher;
  IptelAddress contactaddr = contact;

  debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB",
	"watcher to be stored %s\n",watcher.c_str()) ;
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB",
	"watcher: %s\n wathcer domain: %s\n contact addr user and host %s %s\n",
	(ipaddr.GetUser()).c_str(),(ipaddr.GetHost()).c_str(),
	(contactaddr.GetUser()).c_str(), (contactaddr.GetHost()).c_str());

  int expire_interval = atoi(expires.c_str());
  DBManager *dbm = DBManager::getInstance();
  db_conn_obj *dbc= dbm->getConnectionObj();
  QueryConstructor *qc = new QueryConstructor();
  char* query = (char*)malloc(512);
  memset(query,'\0',512);
	
  qc->update_subscriber_table(query, 512, (char*)to_userdomain.c_str(), 
			      (char*)watcher.c_str(),
			      (char*)event_type.c_str(), expire_interval,
			      (char*)contact.c_str(), (char*)callid.c_str());
	
  debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB",
	"query to update table created %s\n", query);
  long rows = dbm->write_query(dbc, query);
  if(rows <= 0){
    memset(query, '\0', 512);
    qc->insert_subscriber_table(query, 512, (char*)to_userdomain.c_str(), 
				(char*)watcher.c_str(),
				(char*)event_type.c_str(), expire_interval,
				(char*)contact.c_str(), (char*)callid.c_str());
    debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB",
	  "query to insert table created %s\n",query);
    rows = dbm->write_query(dbc, query);
    if(rows >= 1){
      debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB"
	    "successfully created subscription for %s\n", to_userdomain.c_str());
      ret = 1;
    } else {
      
      error("SubsriberHandler::createSubscriptionDB",
	    "failed to insert subscription information in db\n");
      ret = -1;
    }
    
  } else {
    debug(DEBUG_IMPP,"SubscribeEventlistHandler::createSubscriptionDB"
	  "successfully updated subscription for %s, numrows = %d\n", 
	  to_userdomain.c_str(), rows);
    ret = 1;
  }

  free (query);
  delete qc;
  return ret;
}


SubscribeEventlistHandler::~SubscribeEventlistHandler()
{
   debug(DEBUG_IMPP,"SubscribeEventlistHandler::Destructor", 
	 "Default Destructor\n");
}
