#include "subscribe_pres_handler.h"
#include "p_functions.h"
#include "presenceagent.h"
#include "send_notification.h"
#include "send_winfo_notification.h"
#include "send_eventReg.h"
#include "query_constructor.h"
#include "SubscriptionAuthorization.h"
#include "DBManager.h"

/////////////////////////
//Added for RLS
////////////////////////
#include "unsubscription.h"

#define SEND_WINFO 35

SubscribePresHandler::SubscribePresHandler()
{
	debug(DEBUG_IMPP,"SubscribePresHandler::Constructor",
		 "This is the default constructor");
}

SubscribePresHandler::SubscribePresHandler(void* mySIP_pa):SubscribeBaseHandler(mySIP_pa)
{
		
}

SubscribePresHandler::SubscribePresHandler(const SIPRequestMessage & request)
{
	
}

/**
 *	-> Check (in Mmeory (to avoid DB read))
 *	-> if (not in memory)
 *		{ 
 *			-> Check for presentity in DB,[They can be in DB either statically or
 *			Dynamically that is using PUBLISH, this takes care of unconfigured
 *			users]
 *			-> if (not found) {
 *				return 404 User not found
 *			-> else {
 *				-> create the in-memory presentity instance.
 *			}
 *		}
 *		Either we returned or we have in memory record now
 *
 *		-> Check for authorization rules from DB
 *		-> take action based on rules,
 *		-> Send 200, 202, 403 and set Subscription state
 *		-> send NOTIFY as appropriate
 *		-> Create Winfo entry as needed
 *		-> Update subscription for in-memory presentity
 *		-> Update subscription information in (actually in memory) DB for NOTIFY's
 */

SubscribePresHandler::SubscribePresHandler(const SIPRequestMessage & request,
				void*mySIP_pa):SubscribeBaseHandler(mySIP_pa)
{

  int as=-1; // authSub
  int cs=-1; // createSub
  setMethod(METHOD_SUBSCRIBE);
  parseRequest(request);
  
  PresenceFunction *pc=NULL;
  
  // Validate
  int vs =validateSubscription(request);
  if(vs==-1){
    error("SubscribePresHandler::Constructor", 
	  "Invalid headers or body in Subscribe\n");
    //response alread sent in error case,
    // Non error case to be handled ahead.
    return;
  }
  
  //Check if the user is in Memory or DB
  //If the user is valid, then create presentity for 'to_user'
  PresentityDS *pds = checkUserInMemory();
  if(pds==NULL){
    bool ret = checkUserInDB();
    if(ret==false){
      SendResponse(404,"User Not Found");
      return;
    }else {
      debug(DEBUG_IMPP,"SubscribePresHandler::constructor",
	    "User found in DB,creating presentity\n");
      pds =(PresentityDS*)dm->createPresentity(get_to_userdomain());
    }
  }
  setPresentityDS(pds);


  as = authorizeSubscription(request);
  if(as==-1){
    error("SubscribePresHandler::Constructor", 
	  "Internal Server Error in Authorizing Subscription\n");
    SendResponse(500, "Internal Server Error");
    return ;
  }/** else if(as==2){TODO
    // ADD  Winfo, NO_RULE_FOR_SUBSCRIBER,
    // Sent 202, Set state=pending and do normal processing
    // Do we need to send Sub [reg]-> no, what about if rule exists but is not allowed.
    // so DONT SEND EVENT REG if State Pending
    }*/
  
  debug(DEBUG_IMPP,"SubscribePresHandler::constructor","subscriptiopn state %d\n",
	getSubscriptionState());
  

  
   // when do we maintain subscription, check RFC 3857 State machine.    
  cs = createSubscription(request);
  if(cs<1){
    error("SubscribePresHandler::Constructor", 
	  "Internal Server Error in creating  subscription\n");
    SendResponse(500, "Internal Server Error");
    return ;
  }
  
  pc = (PresenceFunction*) new SendPresNotification(this);
  debug(DEBUG_IMPP,"SubscribePresHandler::constructor","SendNotification scheduled\n");
  PresenceHandler::functionList.push_back(pc);
  
  if(sds->getSubscriptionState()!=getSubscriptionState()){
    pc = (PresenceFunction*) new SendWinfoNotification(this);
    debug(DEBUG_IMPP,"SubscribePresHandler::constructor",
	  "SendWinfoNotification scheduled as subscription state changed\n");
    PresenceHandler::functionList.push_back(pc);
  }

  ////////////////////////////
  //Added for RLS
  //////////////////////////
  //Get the 'Expires' header field 
  SIPHeader header_expires = request.GetHeader("Expires");
  string expires = header_expires.GetValue();
  if (expires == "0"){
    pc = (PresenceFunction*) new UnSubscription(this, get_to_userdomain());
    debug(DEBUG_IMPP,"SubscribePresHandler::constructor",
	  "Unsubscription scheduled\n");
    PresenceHandler::functionList.push_back(pc);
  }
  /////////////////////////////
  
  sds->setSubscriptionState(getSubscriptionState());
  
  //RFC 3680 Event Registration package, if event=presence 
  if(EVENT_PRESENCE ==getEvent()){
    //DONT SEND EVENT REG if State Pending TODO
    //pc = (PresenceFunction*) new SendEventReg(this);
    //PresenceHandler::functionList.push_back(pc);
    //debug(DEBUG_IMPP,"SubscribePresHandler::constructor",
    //"event registration function scheduled\n");
    
	}
  debug(DEBUG_IMPP,"SubscribePresHandler::constructor","DONE\n");
	
}

/**
 *	-> Validate type of Event Package
 *	-> Validate value of Expires interval
 *	-> Validate Watcher filter and Store it
 *
*/
int SubscribePresHandler:: validateSubscription(const SIPRequestMessage & request)
{
	SIPHeader header_expire = request.GetHeader("Expires");
	string expires  = header_expire.GetValue(); 

	int expire_interval = 0;
	if(expires!=""){
		expire_interval=atoi((char*)expires.c_str());
		debug(DEBUG_IMPP,
		"SubscribePresHandler::validateSubscription",
		"Expire value is %d\n",expire_interval);
	}else {
		expire_interval=3600;
		error("SubscribePresHandler::validateSubscription",
		"No expires header or header not set,using default value\n");
	}
	SIPHeader header_event = request.GetHeader("Event");
	setEventName(header_event.GetValue());
	if (request.HasBody())
	{
		SIPMessageBody body = request.GetBody();
		setBody(body.GetContent());

#ifdef 	WATCHER_FILTER_OFF
		SendResponse(415,"Application body,Not Understood");
		return -1;
#endif
	}
	
	return 1;
}

int SubscribePresHandler::authorizeSubscription(const SIPRequestMessage & request)
{
	subscription_state_t state_p = unknown;
	debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
	"Authorization for request\n");
	int ret=-1;
	
	/*SubscriptionAuthorization subscription;
        ret = subscription.getAuthorization((char *)get_to_userdomain().c_str(),
		(char*)get_from_userdomain().c_str(),
		(char *)get_from_domain().c_str(),NULL,NULL);
	debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
	"Authorization returned %d \n", ret);
	
	
	setSubscriptionState(unknown);
       */
	
	//TODO: Added for performance testing
	state_p = active;
	setSubscriptionState(active);
	debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
	"Subscription ALLOWED from %s to %s\n",get_from_userdomain().c_str(),
	get_to_userdomain().c_str());
	SendResponse(200,"OK");
	return 1;
	
	if(ret==0){
		/// BLOCKED
		SendResponse(403,"Forbidden");
		debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
		"Subscription BLOCKED for %s by %s\n",get_from_userdomain().c_str(),
		get_to_userdomain().c_str());
		return BLOCKED;
	}else if(ret==1){
		/// Polite Blocked
		setSubscriptionState(pending);
		debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
		"Subscription PENDING for %s by %s\n",get_from_userdomain().c_str(),
		get_to_userdomain().c_str());
		SendResponse(202,"OK");
		return 1;
	}else if(ret==2){
		///Polite Blocked
		state_p =pending;
		setSubscriptionState(pending);
		SendResponse(202,"OK");
		debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
		"Subscription PENDING for %s by %s\n",get_from_userdomain().c_str(),
		get_to_userdomain().c_str());
		return 1;
	}else if(ret==3){
		/// Allowed
		state_p = active;
		setSubscriptionState(active);
		debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
		"Subscription ALLOWED for %s by %s\n",get_from_userdomain().c_str(),
		get_to_userdomain().c_str());
		SendResponse(200,"OK");
		return 1;	
	} 
	if(ret==INTERNAL_DB_ERROR){
		error("SubscribePresHandler::authorizeSubscription", 
		"error occurred in DB querying\n");
	}
	else if(ret==NO_RULE_FOR_SUBSCRIBER){
		state_p =pending;
		setSubscriptionState(pending);
		debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
		"Subscription PENDING for %s by %s, NO RULE FOR SUBSCRIBER\n",
		get_from_userdomain().c_str(),get_to_userdomain().c_str());
		
		SendResponse(202,"OK");
		//returning 2 for winfo
		return 2;
	}
	// Handle Error Response Outside	
	if (ret<0){
		 debug(DEBUG_IMPP,"SubscribePresHandler::authorizeSubscription",
		"Error occurred inAuthorization\n");
	}
	return -1;	
}

int SubscribePresHandler:: createSubscription(const SIPRequestMessage & request)
{
	int ret=1;
	debug(DEBUG_IMPP,"SubscribePresHandler::createSubscription","Entered\n") ;
	bool first_time = false;
	
	PresentityDS *pds = getPresentityDS();
	SubscriberDS *sds = pds->getSubscriberDS(getEvent(),get_from_userdomain());

	if(sds==NULL){
		// This is first subscription
		sds = new SubscriberDS(get_from_userdomain());
                sds->setDialog(getDialog());
		sds->setDomain(get_from_domain());
		char* rn= (char*)malloc(4);
		memset(rn,'\0',4);
		sprintf(rn,"%d",(rand()%100));
		string random_num =rn;
		sds->setId(string(get_to_user())+random_num);
		free(rn);
		
		sds->set_subsriber_contact(getEvent(),get_contact_addr());
		first_time =true;
		
		sds->setSubscriptionState(unknown);
		
		pds->store(getEvent(),get_from_userdomain(),sds);
		debug(DEBUG_IMPP,"SubscribePresHandler:createSubscription",
		"Subscriber %s created and added for the presentity %s \n",
		(get_from_userdomain()).c_str(),(get_to_userdomain()).c_str()) ;
	}else{
		//TODO : update subscription dialog information
		sds->set_subsriber_contact(getEvent(),get_contact_addr());
		debug(DEBUG_IMPP,"SubscribePresHandler:createSubscription",
		"subscriber %s information updated for presentity %s\n",
		(get_from_userdomain()).c_str(),(get_to_userdomain()).c_str()) ;
	}
	setSubscriberDS(sds);
	
	AugmentedDOMNode *adn = (AugmentedDOMNode*)sds->getData(getEvent());
	if(adn==NULL){
		first_time=true;
	}
	if(first_time==true) // CONDITIONAL INSERT **
	{
		ret = createSubscriptionDB(request);
	}
	else {
		// TODO: Update expiration info and dialog info
	}
	debug(DEBUG_IMPP,"SubscribePresHandler::createSubscription","Done \n");
	return ret;
}

int SubscribePresHandler :: createSubscriptionDB(const SIPRequestMessage & request)
{
	int ret =-1;

	debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB",
		"New subscription by %s\n",(get_from_user()).c_str());
	SIPHeader header_from = request.GetHeader("From");
	SIPHeader header_to = request.GetHeader("To");
	SIPHeader header_contact = request.GetHeader("Contact");
	SIPHeader header_event = request.GetHeader("Event");
	SIPHeader header_expires = request.GetHeader("Expires");
	SIPHeader header_callid = request.GetHeader("Call-ID");
	string watcher = header_from.AsString(); 
	string presentity = header_to.AsString(); 
	string contact  = header_contact.GetValue(); 
	string event_type = header_event.GetValue();
	string expires = header_expires.GetValue();
	string callid = header_expires.GetValue();
	IptelAddress ipaddr = watcher;
	IptelAddress contactaddr = contact;

	debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB",
		"watcher to be stored %s\n",watcher.c_str()) ;
	debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB",
	 "watcher is %s  \n wathcer domain is %s\n contact addr user and host %s %s\n",
	(ipaddr.GetUser()).c_str(),(ipaddr.GetHost()).c_str(),
	 (contactaddr.GetUser()).c_str(), (contactaddr.GetHost()).c_str());

	int expire_interval = atoi(expires.c_str());
 	DBManager *dbm = DBManager::getInstance();
	db_conn_obj *dbc= dbm->getConnectionObj();
	QueryConstructor *qc = new QueryConstructor();
	char* query = (char*)malloc(512);
	memset(query,'\0',512);
	
	qc->update_subscriber_table(query,512,(char*)presentity.c_str(),(char*)watcher.c_str(),
	(char*)event_type.c_str(),expire_interval,(char*)contact.c_str(),(char*)callid.c_str());
	
	debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB",
		 "query to update table created %s\n",query);
	long rows= dbm->write_query(dbc,query);
	if(rows<=0){
		memset(query,'\0',512);
		qc->insert_subscriber_table(query,512,(char*)presentity.c_str(),(char*)watcher.c_str(),
		(char*)event_type.c_str(),expire_interval,(char*)contact.c_str(),(char*)callid.c_str());
		debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB",
		 "query to insert table created %s\n",query);
		rows= dbm->write_query(dbc,query);
		if(rows>=1){
			debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB"
		 	"successfully created subscription for %s\n",
			 presentity.c_str());
                         ret =1;
		} else {

			error("SubsriberHandler::createSubscriptionDB",
			"failed to insert subscription information in db\n");
			ret =  -1;
		}
	
	} else {
			debug(DEBUG_IMPP,"SubscribePresHandler::createSubscriptionDB"
		  	"successfully updated subscription for %s,numrows=%d\n",
			 presentity.c_str(),rows);
                        ret =1;
	}
        free (query);
        delete qc;
	return ret;
}

SubscribePresHandler:: ~SubscribePresHandler()
{
   debug(DEBUG_IMPP,"SubscribePresHandler::Destructor", "Default Destructor\n");
}
